from math import radians

try:
    from typing import Self
except ImportError:
    # py3.10 and lower
    from typing_extensions import Self

import raimad as rai

class AnSecError(Exception):
    pass

class AnSecRadiusError(AnSecError):
    pass

class AnSecRadiusTooManyArgumentsError(AnSecRadiusError):
    pass

class AnSecRadiusNotEnoughArgumentsError(AnSecRadiusError):
    pass

class AnSecRadiusIncorrectArgumentsError(AnSecRadiusError):
    pass

class AnSecThetaError(AnSecError):
    pass

class AnSecThetaTooManyArgumentsError(AnSecThetaError):
    pass

class AnSecThetaNotEnoughArgumentsError(AnSecThetaError):
    pass

class AnSecThetaIncorrectArgumentsError(AnSecThetaError):
    pass

class AnSec(rai.Compo):
    """
    Annular Sector

    A polygon approximating an annular sector,
    or, in working men's terms, a "pizza crust".
    There are many possible ways to define an AnSec;
    # TODO examples
    """

    browser_tags = ["builtin", "polygon"]

    class Options:
        r1 = rai.Option("Inner radius", browser_default=18)
        r2 = rai.Option("Outter radius", browser_default=20)
        theta1 = rai.Option("Angle 1", browser_default=radians(45))
        theta2 = rai.Option("Angle 2", browser_default=radians(90))

    def _make(
            self,
            r1: float,
            r2: float,
            theta1: float,
            theta2: float,
            num_points: int = 100,
            ) -> None:

        if abs(r1 - r2) < rai.epsilon:
            self.geoms.update({'root': [[]]})
            return

        theta_step = (theta2 - theta1) / num_points
        angles = tuple(theta1 + theta_step * i for i in range(num_points))

        self.geoms.update({
            'root': [
                [
                    rai.polar(arg=angle, mod=radius)
                    for radius, angles in (
                        (r1, angles),
                        (r2, angles[::-1]),
                        # Fun exercise: change the above two tuples to lists
                        # and see what mypy has to say about that
                        )
                    for angle in angles
                    ]
                ]
            })

    @classmethod
    def from_auto(
            cls,
            r1: float | None = None,
            r2: float | None = None,
            rmid: float | None = None,
            dr: float | None = None,
            theta1: float | None = None,
            theta2: float | None = None,
            thetamid: float | None = None,
            dtheta: float | None = None,
            num_points: int = 100,
            ) -> Self:

        r1, r2 = cls.interpret_radius(r1, r2, rmid, dr)
        theta1, theta2 = cls.interpret_theta(
            theta1,
            theta2,
            thetamid,
            dtheta,
            )

        return cls(r1, r2, theta1, theta2, num_points)

    @staticmethod
    def interpret_radius(
            r1: float | None,
            r2: float | None,
            rmid: float | None,
            dr: float | None,
            ) -> tuple[float, float]:

        if r1 is None and r2 is None and rmid is None and dr is None:
            raise AnSecRadiusNotEnoughArgumentsError(
                "No radius specified"
                )

        if r1 is None and r2 is None and rmid is None and dr is not None:
            raise AnSecRadiusNotEnoughArgumentsError(
                "Radius delta specified, but no r1 or r2"
                )

        if r1 is None and r2 is None and rmid is not None and dr is None:
            raise AnSecRadiusNotEnoughArgumentsError(
                "Midradius specified, but no radius delta"
                )

        if r1 is None and r2 is None and rmid is not None and dr is not None:
            return rmid - dr / 2, rmid + dr / 2

        if r1 is None and r2 is not None and rmid is None and dr is None:
            raise AnSecRadiusNotEnoughArgumentsError(
                "r2 specified, but no r1 or radius delta"
                )

        if r1 is None and r2 is not None and rmid is None and dr is not None:
            return r2 - dr, r2

        if r1 is None and r2 is not None and rmid is not None and dr is None:
            raise AnSecRadiusIncorrectArgumentsError(
                "r2 and midradius specified. "
                "What do you want me to do??"
                )

        if r1 is None and r2 is not None and rmid is not None and \
                dr is not None:
            raise AnSecRadiusTooManyArgumentsError(
                "r2, midradius, and radius delta specified. "
                "You need only two of these!"
                )

        if r1 is not None and r2 is None and rmid is None and dr is None:
            raise AnSecRadiusNotEnoughArgumentsError(
                "Only r1 specified. "
                "Specify radius delta or r2."
                )

        if r1 is not None and r2 is None and rmid is None and dr is not None:
            return r1, r1 + dr

        if r1 is not None and r2 is None and rmid is not None and dr is None:
            raise AnSecRadiusIncorrectArgumentsError(
                "R1 and midradius specified. "
                "What do you want me to do??"
                )

        if r1 is not None and r2 is None and rmid is not None and \
                dr is not None:
            raise AnSecRadiusTooManyArgumentsError(
                "r1, midradius, and radius delta specified. "
                "You need only two of these!"
                )

        if r1 is not None and r2 is not None and rmid is None and dr is None:
            return r1, r2

        if r1 is not None and r2 is not None and rmid is None and dr \
                is not None:
            raise AnSecRadiusTooManyArgumentsError(
                "r1, r2, and radius delta specified. "
                "What do you want me to do??"
                )

        if r1 is not None and r2 is not None and rmid is not None and \
                dr is None:
            raise AnSecRadiusTooManyArgumentsError(
                "r1, r2, and midradius specified. "
                "What do you want me to do??"
                )

        raise AnSecRadiusTooManyArgumentsError(
                "What do you want me to do??"
                )

    @staticmethod
    def interpret_theta(
            t1: float | None,
            t2: float | None,
            tmid: float | None,
            dt: float | None,
            ) -> tuple[float, float]:

        if t1 is None and t2 is None and tmid is None and dt is None:
            raise AnSecThetaNotEnoughArgumentsError(
                "No Theta specified"
                )

        if t1 is None and t2 is None and tmid is None and dt is not None:
            raise AnSecThetaNotEnoughArgumentsError(
                "Theta delta specified, but no t1 or t2"
                )

        if t1 is None and t2 is None and tmid is not None and dt is None:
            raise AnSecThetaNotEnoughArgumentsError(
                "MidTheta specified, but no Theta delta"
                )

        if t1 is None and t2 is None and tmid is not None and dt is not None:
            return tmid - dt / 2, tmid + dt / 2

        if t1 is None and t2 is not None and tmid is None and dt is None:
            raise AnSecThetaNotEnoughArgumentsError(
                "t2 specified, but no t1 or Theta delta"
                )

        if t1 is None and t2 is not None and tmid is None and dt is not None:
            return t2 - dt, t2

        if t1 is None and t2 is not None and tmid is not None and dt is None:
            raise AnSecThetaIncorrectArgumentsError(
                "t2 and midTheta specified. "
                "What do you want me to do??"
                )

        if t1 is None and t2 is not None and tmid is not None and \
                dt is not None:
            raise AnSecThetaTooManyArgumentsError(
                "t2, midTheta, and Theta delta specified. "
                "You need only two of these!"
                )

        if t1 is not None and t2 is None and tmid is None and dt is None:
            raise AnSecThetaNotEnoughArgumentsError(
                "Only t1 specified. "
                "Specify Theta delta or t2."
                )

        if t1 is not None and t2 is None and tmid is None and dt is not None:
            return t1, t1 + dt

        if t1 is not None and t2 is None and tmid is not None and dt is None:
            raise AnSecThetaIncorrectArgumentsError(
                "t1 and midTheta specified. "
                "What do you want me to do??"
                )

        if t1 is not None and t2 is None and tmid is not None and \
                dt is not None:
            raise AnSecThetaTooManyArgumentsError(
                "t1, midTheta, and Theta delta specified. "
                "You need only two of these!"
                )

        if t1 is not None and t2 is not None and tmid is None and dt is None:
            return t1, t2

        if t1 is not None and t2 is not None and tmid is None and \
                dt is not None:
            raise AnSecThetaTooManyArgumentsError(
                "t1, t2, and Theta delta specified. "
                "What do you want me to do??"
                )

        if t1 is not None and t2 is not None and tmid is not None and \
                dt is None:
            raise AnSecThetaTooManyArgumentsError(
                "t1, t2, and midTheta specified. "
                "What do you want me to do??"
                )

        raise AnSecThetaTooManyArgumentsError(
                "What do you want me to do??"
                )




import unittest

import raimad as rai

from .utils import GeomsEqual, ArrayAlmostEqual

class TestPolys(GeomsEqual, ArrayAlmostEqual, unittest.TestCase):

    def test_rectlw(self):
        rect = rai.RectLW(10, 20)
        self.assertGeomsEqual(
            rect.geoms,
            {
                'root': [
                    [
                        [-5, -10],
                        [5, -10],
                        [5, 10],
                        [-5, 10]
                        ],
                    ]
                }
            )

    def test_circle(self):
        circle = rai.Circle(69)
        points = circle.geoms['root'][0]
        self.assertGreaterEqual(len(points), 10)
        for point in points:
            self.assertAlmostEqual(
                rai.affine.norm(point),
                69
                )

    def test_custompoly(self):
        """
        Test CustomPoly
        """
        poly = rai.CustomPoly([
            [10, 10],
            [20, 10],
            [20, 30],
            ])
        self.assertGeomsEqual(
            poly.geoms,
            {
                'root': [
                    [
                        [10, 10],
                        [20, 10],
                        [20, 30],
                        ],
                    ]
                }
            )

    def test_custompoly_marks(self):
        """
        Test CustomPoly with dynamic marks
        """
        poly = rai.CustomPoly([
            ['start', [10, 10]],
            [20, 10],
            ['end', [20, 30]],
            ])
        self.assertGeomsEqual(
            poly.geoms,
            {
                'root': [
                    [
                        [10, 10],
                        [20, 10],
                        [20, 30],
                        ],
                    ]
                }
            )
        self.assertEqual(
            poly.marks.start,
            (10, 10)
            )
        self.assertEqual(
            poly.marks.end,
            (20, 30)
            )


if __name__ == '__main__':
    unittest.main()




"""svg.py: tools for rendering compos as svg pictures."""

from typing import Iterator

import raimad as rai

def export_svg(compo: 'rai.typing.CompoLike') -> str:
    """
    Make an SVG representation of a component.

    Parameters
    ----------
    compo : rai.typing.CompoLike
        Component (or Proxy pointing to a component)
        to export to svg.

    Returns
    -------
    str
        A string containing the SVG image.
    """
    return ''.join(yield_svg(compo))

def yield_svg(compo: 'rai.typing.CompoLike') -> Iterator[str]:
    """
    Make an SVG representation of a component.

    Unlike `export_svg`, this function yields parts of the SVG
    file, instead of returning the entire string all at once.

    Parameters
    ----------
    compo : rai.typing.CompoLike
        Component (or Proxy pointing to a component)
        to export to svg.

    Returns
    -------
    Iterator[str]
        strings containing fragments of the SVG file
    """
    bbox = compo.bbox.pad(10)

    yield (
        '<svg xmlns="http://www.w3.org/2000/svg" '
        f'width="{bbox.length}" '
        f'height="{bbox.width}" '
        '>\n'
        '\n'
        f'<g transform="translate({-bbox.left},{bbox.top})">\n'
        # TODO handle neg left correctly?
        '<g transform="scale(1,-1)">\n'
        )

    for layer_name, layer_geoms in compo.steamroll().items():
        for geom in layer_geoms:
            yield (
                '<polygon '
                'fill="#00000000" '
                'stroke="#000000" '
                'stroke-width="1" '
                'points="'
                )

            for point in geom:
                yield f'{point[0]},{point[1]} '

            yield '" />\n'

    yield '</g></g></svg>\n'




import unittest

import raimad as rai

class BareGeometric(rai.Compo):
    def _make(self):
        self.geoms.update({
            'root': [
                [
                    (0, 0),
                    (0, 10),
                    (10, 10),
                    (10, 0),
                    ],
                [
                    (20, 20),
                    (40, 20),
                    (30, 40),
                    ],
                ]
            })

class BareStructural(rai.Compo):
    def _make(self):
        self.subcompos.append(
            rai.Proxy(
                BareGeometric(),
                {'root': 'upper'},
                rai.Transform().scale(2)
                )
            )
        self.subcompos.append(
            rai.Proxy(
                BareGeometric(),
                {'root': 'lower'},
                rai.Transform().scale(0.5)
                )
            )

class TestCompo(unittest.TestCase):

    def test_transform(self):
        compo = BareStructural()
        geoms = compo.steamroll()

        # Test one root layer
        self.assertEqual(geoms.keys(), {'upper', 'lower'})

        # Test existence of polys
        self.assertEqual(len(geoms['upper']), 2)
        self.assertEqual(len(geoms['lower']), 2)

        # Test location
        self.assertEqual(list(geoms['upper'][0][2]), [20, 20])
        self.assertEqual(list(geoms['lower'][0][2]), [5, 5])

    def test_proxy_transform(self):
        compo = BareStructural()

        geoms_0 = compo.subcompos[0].steamroll()
        geoms_1 = compo.subcompos[1].steamroll()

        self.assertEqual(geoms_0.keys(), {'upper'})
        self.assertEqual(geoms_1.keys(), {'lower'})

        # Test existence of polys
        self.assertEqual(len(geoms_0['upper']), 2)
        self.assertEqual(len(geoms_1['lower']), 2)

        # Test location
        self.assertEqual(list(geoms_0['upper'][0][2]), [20, 20])
        self.assertEqual(list(geoms_1['lower'][0][2]), [5, 5])

    def test_walk_hier(self):
        mycompo = BareStructural()
        hier = list(mycompo.walk_hier())

        self.assertEqual(sum((compo is mycompo for compo in hier)), 1)
        self.assertEqual(
            sum(
                compo.final() is mycompo.subcompos[0].final()
                for compo in hier
                ),
            1)
        self.assertEqual(
            sum(
                compo.final() is mycompo.subcompos[1].final()
                for compo in hier
                ),
            1)

        self.assertEqual(sum((compo.depth() == 0 for compo in hier)), 1)
        self.assertEqual(sum((compo.depth() == 1 for compo in hier)), 2)


if __name__ == '__main__':
    unittest.main()




"""noreuse.py: home to the NoReuse CIF exporter."""

from typing import Iterator

import raimad as rai

class NoReuse:
    """CIF Exporter that doesn't reuse subroutines."""

    def __init__(
            self,
            compo: 'rai.typing.CompoLike',
            multiplier: float = 1e3
            ) -> None:

        self.compo = compo
        self.rout_num = 1
        self.multiplier = multiplier

        self.cif_string = self._export_cif()

    def _export_cif(self) -> str:
        return ''.join(self._yield_cif())

    def _yield_cif(self) -> Iterator[str]:
        """Yield lines of cif file."""
        first_rout = self.rout_num
        yield from self.yield_cif_bare(self.compo)
        yield f'C {first_rout};\n'
        yield 'E'

    def yield_cif_bare(self, compo: 'rai.typing.CompoLike') -> Iterator[str]:
        """Yield lines of CIF of a particular component, without calling it."""
        # Opening line, define the routine
        yield f'DS {self.rout_num} 1 1;\n'

        # advance to next routine number
        self.rout_num += 1

        # Export all geometries
        for layer, geom in compo.geoms.items():
            yield f'\tL L{layer};\n'
            for poly in geom:
                yield '\tP '
                for point in poly:
                    yield (
                        f'{int(point[0] * self.multiplier)} '
                        f'{int(point[1] * self.multiplier)} '
                        )
                yield ';\n'

        # Precompute a list of [routine number, subcomponent]
        # Remember, subcomponents can also have subcomponents,
        # so the subroutine numbers won't always be consecutive.
        # There's no way to know ahead of time,
        # you just have to generate each subcompo and keep track of
        # the routine number
        subcompos = []
        for subcompo in compo.subcompos.values():
            subcompos.append((
                self.rout_num,
                list(self.yield_cif_bare(subcompo))
                ))

        # Call subcomponent procedures
        for i, _ in subcompos:
            yield f'\tC {i};\n'
        yield 'DF;\n'

        # Define those procedures
        for _, this_subcompo in subcompos:
            yield from this_subcompo




"""
Autotests for dictlist.
"""

import unittest

from raimad.dictlist import DictList

class TestDictList(unittest.TestCase):

    def test_dictlist_as_dict(self):
        dl = DictList()
        dl["a"] = 5
        dl["b"] = 6

        self.assertEqual(len(dl), 2)
        self.assertEqual(set(dl.keys()), {"a", "b"})
        self.assertEqual(set(dl.values()), {5, 6})
        self.assertEqual(dl["a"], 5)
        self.assertEqual(dl["b"], 6)

        dl["a"] += 2

        self.assertEqual(len(dl), 2)
        self.assertEqual(set(dl.keys()), {"a", "b"})
        self.assertEqual(set(dl.values()), {7, 6})
        self.assertEqual(dl["a"], 7)
        self.assertEqual(dl["b"], 6)

    def test_dictlist_as_list(self):
        dl = DictList()
        dl.append(4)
        dl.append(5)

        self.assertEqual(len(dl), 2)
        self.assertEqual(set(dl.keys()), {0, 1})
        self.assertEqual(set(dl.values()), {4, 5})
        self.assertEqual(dl[0], 4)
        self.assertEqual(dl[1], 5)

        dl[1] = 8

        self.assertEqual(len(dl), 2)
        self.assertEqual(dl[0], 4)
        self.assertEqual(dl[1], 8)
        self.assertEqual(set(dl.keys()), {0, 1})
        self.assertEqual(set(dl.values()), {4, 8})

    def test_dictlist_as_object(self):
        dl = DictList()
        dl.ayy = 5
        dl.b = 6

        self.assertEqual(len(dl), 2)
        self.assertEqual(set(dl.keys()), {"ayy", "b"})
        self.assertEqual(set(dl.values()), {5, 6})
        self.assertEqual(dl.ayy, 5)
        self.assertEqual(dl.b, 6)

        dl.ayy += 2

        self.assertEqual(len(dl), 2)
        self.assertEqual(set(dl.keys()), {"ayy", "b"})
        self.assertEqual(set(dl.values()), {7, 6})
        self.assertEqual(dl.ayy, 7)
        self.assertEqual(dl.b, 6)

    def test_dictlist_indexing(self):
        dl = DictList()
        dl["a"] = 5
        dl.b = 6

        self.assertEqual(len(dl), 2)
        self.assertEqual(set(dl.keys()), {"a", "b"})
        self.assertEqual(set(dl.values()), {5, 6})
        self.assertEqual(dl["a"], 5)
        self.assertEqual(dl["b"], 6)
        self.assertEqual(dl[0], 5)
        self.assertEqual(dl[1], 6)

        with self.assertRaises(IndexError):
            dl[2]

        with self.assertRaises(IndexError):
            dl[2] = 2

        dl[1] += 2

        self.assertEqual(len(dl), 2)
        self.assertEqual(set(dl.keys()), {"a", "b"})
        self.assertEqual(set(dl.values()), {5, 8})
        self.assertEqual(dl["a"], 5)
        self.assertEqual(dl["b"], 8)
        self.assertEqual(dl[0], 5)
        self.assertEqual(dl[1], 8)

    def test_dictlist_assign(self):
        dl = DictList()

        with self.assertRaises(KeyError):
            dl["keys"] = 5

        with self.assertRaises(KeyError):
            dl["_mydict"] = 5

        dl._mything = 5
        self.assertEqual(dl._mything, 5)
        self.assertEqual(len(dl), 0)

        with self.assertRaises(KeyError):
            dl["_mything"]

        with self.assertRaises(AttributeError):
            dl.keys = 5

    def _test_dictlist_filter(self, dl):
        """
        Helper method for testing whether _filter_set or _filter_get
        methods work correctly.
        They work exactly the same, but at different points in time
        (assigning an item vs querying and item),
        so we can use the same test payload.
        """

        dl.append('ayy')
        dl['foo'] = 'bar'
        dl.baz = 'baq'

        self.assertEqual(dl[0], 'AYY')
        self.assertEqual(dl[1], 'BAR')
        self.assertEqual(dl[2], 'BAQ')

        self.assertEqual(dl['foo'], 'BAR')
        self.assertEqual(dl['baz'], 'BAQ')

        self.assertEqual(dl.foo, 'BAR')
        self.assertEqual(dl.baz, 'BAQ')

        self.assertEqual(set(dl.keys()), {0, 'foo', 'baz'})
        self.assertEqual(set(dl.values()), {'AYY', 'BAR', 'BAQ'})
        self.assertEqual(
            [[key, val] for key, val in dl.items()],
            [
                [0, 'AYY'],
                ['foo', 'BAR'],
                ['baz', 'BAQ'],
                ]
            )

    def test_dictlist_filter_set(self):
        class DictListSet(DictList):
            def _filter_set(self, val):
                return val.upper()

        self._test_dictlist_filter(DictListSet())

    def test_dictlist_filter_get(self):
        class DictListGet(DictList):
            def _filter_get(self, val):
                return val.upper()

        self._test_dictlist_filter(DictListGet())

    def test_dictlist_init(self):
        class DictListSet(DictList):
            def _filter_set(self, val):
                return val.upper()

        dl = DictList()
        dl.append('ayy')
        dl_view = DictListSet(dl._dict, copy=False)
        dl_view.append('lmao')

        self.assertEqual(dl[0], 'ayy')
        self.assertEqual(dl[1], 'LMAO')

        with self.assertRaises(TypeError):
            DictList(dl._dict)

        with self.assertRaises(TypeError):
            DictList(dl._dict, True)

        with self.assertRaises(TypeError):
            DictList(dl)

        with self.assertRaises(TypeError):
            DictList('invalid type')

    def test_dictlist_post_init(self):
        class DictListPostInit(DictList):
            def _post_init(self):
                self._myattr = 'hello'

        dl = DictListPostInit()
        self.assertEqual(dl._myattr, 'hello')


if __name__ == '__main__':
    unittest.main()




"""__init__.py: Namespace flattening for RAIMAD."""

import sys

epsilon = sys.float_info.epsilon

from raimad.empty import Empty, EmptyType
from raimad import graphviz as gv
from raimad.helpers import (
    fullcircle,
    halfcircle,
    quartercircle,
    eigthcircle,
    semicircle,
    demisemicircle,
    hemidemisemicircle,
    angle_between,
    polar,
    is_compo_class,
    custom_base,
    WINGDINGS,
    wingdingify,
    midpoint,
    add,
    eq,
    sub,
    )
from raimad import affine
import raimad.iters as iters
from raimad.iters import (
    overlap,
    nonoverlap,

    duplets,
    triplets,
    quadlets,
    quintlets,

    couples,
    triples,
    quadles,
    quintles,

    flatten,
    braid
    )
from raimad.string_import import string_import
from raimad.docparse import split_docstring

from raimad.dictlist import FilteredDictList, DictList

from raimad.annotation import Annotation
from raimad.mark import Mark
from raimad.layer import Layer
from raimad.option import Option
from raimad.boundpoint import BoundPoint
from raimad.transform import Transform
from raimad.compo import Compo
from raimad.compo import MarksContainer
from raimad.compo import SubcompoContainer
from raimad.proxy import Proxy
from raimad.proxy import LMap
from raimad.partial import Partial
from raimad.bbox import AbstractBBox, BBox
from raimad.boundbbox import BoundBBox

from raimad.rectlw import RectLW
from raimad.rectwire import RectWire
from raimad.circle import Circle
from raimad.ansec import AnSec
from raimad.custompoly import CustomPoly

from raimad import typing
from raimad import typing as t

from raimad import cif
from raimad.cif.shorthand import export_cif
from raimad.svg import export_svg
from raimad.show import show
from raimad import err

from raimad.snowman import Snowman

# The function of __all__ is to specify which things get imported
# when someone does `from raimad import *`,
# but it also lets tools like ruff and mypy know that we're
# importing these things for the purpose of namespace flattening,
# so they don't complain that these are "unused" imports
__all__ = [
    'export_svg',
    'Transform',
    'Compo',
    'FilteredDictList',
    'DictList',
    'Proxy',
    'BoundPoint',
    'MarksContainer',
    'SubcompoContainer',
    'AbstractBBox',
    'BBox',
    'BoundBBox',
    'Partial',
    'LMap',
    'export_cif',
    'string_import',
    'Annotation',
    'Mark',
    'Layer',
    'Option',
    'Empty',
    'EmptyType',
    't',
    'typing',
    'AnSec',
    'Circle',
    'RectLW',
    'RectWire',
    'CustomPoly',
    'Snowman',
    'err',
    'cif',
    'split_docstring',  # TODO is needed?
    'braid',
    'flatten',
    'quintles',
    'quadles',
    'triples',
    'couples',
    'quintlets',
    'quadlets',
    'triplets',
    'duplets',
    'overlap',
    'nonoverlap',
    'iters',
    'affine',
    'fullcircle',
    'halfcircle',
    'quartercircle',
    'eigthcircle',
    'semicircle',
    'demisemicircle',
    'hemidemisemicircle',
    'angle_between',
    'polar',
    'is_compo_class',
    'custom_base',
    'WINGDINGS',
    'wingdingify',
    'midpoint',
    'gv',
    'add',
    'eq',
    'epsilon',
    'show'
    ]





"""show.py: show() function and various related cross-platform hackery."""

import tempfile
import platform
import os
import subprocess
import shutil
import shlex
from pathlib import Path

import sys

import raimad as rai

def is_native_klayout_installed() -> bool:
    """Check whether KLayout is installed natively (Linux)."""
    return bool(shutil.which('klayout'))

def is_klayout_installed_win() -> bool:
    """Check whether KLayout is installed (Windows)."""
    appdata = os.getenv('APPDATA')
    return Path(rf"{appdata}\KLayout\klayout_app.exe").exists()

def is_flatpak_installed() -> bool:
    """Check whether Flatpak is installed (Linux)."""
    return bool(shutil.which('flatpak'))

def is_flatpak_klayout_installed() -> bool:
    """Checked whether KLayout is installed via flatpak (Linux)."""
    result = subprocess.run(
        ["flatpak", "list", "--app", "--columns=application"],
        capture_output=True,
        text=True,
        check=True
    )
    return "de.klayout.KLayout" in result.stdout

def get_custom_cifview_command() -> str | None:
    """Check if an env variable set that specifies a custom CIF viewer."""
    return (
        os.environ.get("RAIMAD_CIF_VIEWER")
        or
        os.environ.get("CIF_VIEWER")
        or
        None
        )

def get_klayout_app_mac() -> str | None:
    """Get path to klayout.app on Mac OS, None if not installed."""
    result = subprocess.run(
        [
            "sh",
            "-c",
            """mdfind "kMDItemKind == 'Application'" | grep klayout"""
            ],
        capture_output=True,
        text=True,
        check=True
        ).stdout.strip()

    if result:
        return result
    return None

def is_klayout_running() -> bool:
    """Check whether klayout is already running."""

    if platform.system() in {"Linux", "Darwin"}:
        return subprocess.run(
            ["ps", "aux"],
            capture_output=True,
            text=True,
            check=True
            ).stdout.count("klayout") > 0

    elif platform.system() == "Windows":
        return subprocess.run(
            ["tasklist"],
            capture_output=True,
            text=True,
            check=True
            ).stdout.count("klayout_app.exe") > 0

    raise NotImplementedError(
        "raimad.show() is not available on your platform yet. "
        "Contact maybetree and request support."
        )


def get_cifview_args(file: str) -> tuple[str, ...]:
    """Get a command to open `file` in a CIF viewer."""

    custom_command = get_custom_cifview_command()
    if custom_command:
        return (
            *(
                file if arg == '__FILE__' else arg
                for arg in
                shlex.split(custom_command)
                ),
            *((file, ) * ('__FILE__' not in custom_command))
            )

    if platform.system() == "Linux":
        if is_native_klayout_installed():
            return ("klayout", file)

        if is_flatpak_installed():
            if is_flatpak_klayout_installed():
                return (
                    "flatpak",
                    "run",
                    "--file-forwarding",
                    "de.klayout.KLayout",
                    "@@",
                    file,
                    "@@"
                    )

        raise Exception(
            'I could not figure out how to show you the CIF file. '
            'Please install KLayout using your system package manager '
            'or flatpak. '
            'To use a different CIF viewer '
            '(or KLayout installed at a custom path) '
            'set the following environment variable: '
            'CIF_VIEWER="your_viewer_program __FILE__". '
            )

    elif platform.system() == "Windows":
        if is_klayout_installed_win():
            appdata = os.getenv('APPDATA')
            return (
                rf"{appdata}\KLayout\klayout_app.exe",
                file,
                )

        raise Exception(
            'I could not figure out how to show you the CIF file. '
            'Please install KLayout. '
            'Is KLayout already installed at a non-standard path, '
            'or would you like to use a different CIF viewer? '
            'Then set the following environment variable: '
            'CIF_VIEWER="your_viewer_program __FILE__". '
            )

    elif platform.system() == "Darwin":
        app = get_klayout_app_mac()
        if app:
            return (
                f"{app}/Contents/MacOS/klayout",
                file
                )

        raise Exception(
            'I could not figure out how to show you the CIF file. '
            'Please install KLayout. '
            'Is KLayout already installed at a non-standard path, '
            'or would you like to use a different CIF viewer? '
            'Then set the following environment variable: '
            'CIF_VIEWER="your_viewer_program __FILE__". '
            )

    raise NotImplementedError(
        "raimad.show() is not available on your platform yet. "
        "Contact maybetree and request support."
        )


def show(compo: 'rai.typing.CompoLike', ignore_running: bool = False) -> None:
    """Export `compo` and open it in a CIF viewer."""

    file = Path(tempfile.gettempdir()) / "RAIMAD-SHOW.cif"
    rai.export_cif(compo, file)
    print(f"Saved to {file}")

    if not ignore_running and is_klayout_running():
        print("Klayout already running.")
        return

    args = get_cifview_args(str(file))
    print(f"Running {args}...")

    subprocess.Popen(args)




"""custompoly.py: home to CustomPoly builtin component."""
import raimad as rai

from typing import Sequence

class CustomPolyException(Exception):
    """Error for when you incorrectly create a CustomPoly."""

class CustomPoly(rai.Compo):
    """
    A polygon that you can defined on-the fly.

    #TODO add example
    """

    browser_tags = ["builtin", "polygon"]

    class Options:
        points = rai.Option(
            "A list of points for the polygon",
            browser_default=[
                [0, 0],
                [10, 0],
                [10, 20],
                [20, -20],
                ]
            )

    def _make(
            self,
            points: Sequence[
                tuple[float, float]
                |
                tuple[str, tuple[float, float]]
                ]
            ) -> None:
        processed_points: 'list[rai.typing.Point]' = []
        for point in points:

            if isinstance(point[0], str):
                self.marks[point[0]] = point[1]
                processed_points.append(point[1])

            else:
                processed_points.append(point)

        self.geoms.update({'root': [processed_points]})




"""
Autotests for boundpoint.
A boundpoint is a point tied to a proxy.
They can be used for performing transformations
relative to a specific point of a compo.
They are used in marks and bboxes.
For example, `mycompo.bbox.mid.to(0, 0)`
moves `mycompo` so that its middle (center of bounding box) is at the orign.
`mycompo.marks.coupler.hflip()` flips `mycompo` horizontally
around the mark called `coupler`.
"""

import unittest

import raimad as rai

from .utils import ArrayAlmostEqual

class BareGeometric(rai.Compo):
    def _make(self):
        self.geoms.update({
            'root': [
                [
                    [0, 0],
                    [40, 0],
                    [0, 40],
                    ],
                ]
            })
        # These are the two different syntaxes for assigning marks,
        # we want to test them both
        self.marks.corner_a = (0, 0)
        self.marks['corner_b'] = (0, 40)

class BareStructural(rai.Compo):
    def _make(self):
        self.subcompos.geometric = BareGeometric().proxy().move(3, 2)
        self.marks.corner_a = self.subcompos.geometric.marks.corner_a

class TestBoundpoint(ArrayAlmostEqual, unittest.TestCase):

    def test_boundpoint_mark(self):
        """
        Test that getting the mark on a compo gives
        the correct point
        """
        compo = BareGeometric()

        self.assertArrayAlmostEqual(
            compo.marks.corner_a,
            (0, 0)
            )

    def test_boundpoint_mark_translated(self):
        """
        Test that getting the mark on a proxy
        gives the correct point in external coordinates
        by applying the transform
        """
        compo = BareGeometric().proxy()
        compo.move(2, 3)

        self.assertArrayAlmostEqual(
            compo.marks["corner_a"],
            (2, 3)
            )

    def test_boundpoint_mark_subcompo(self):
        """
        Test that getting the mark from a subcompo
        still computes the external coordinates correctly
        """
        compo = BareStructural()

        self.assertArrayAlmostEqual(
            compo.subcompos.geometric.marks.corner_a,
            (3, 2)
            )

    def test_boundpoint_mark_subcompo_translated(self):
        """
        Test that getting the mark from a subcompo of a proxy
        still computes the external coordinates correctly
        (i.e. applies the transform done inside the structural
        componenet, and the transform done by the proxy created this function)
        """
        compo = BareStructural().proxy()
        compo.move(10, 20)

        self.assertArrayAlmostEqual(
            compo.subcompos.geometric.marks.corner_a,
            (3 + 10, 2 + 20)
            )

    def test_boundpoint_bbox_subcompo_translated(self):
        compo = BareStructural().proxy()
        compo.move(10, 20)

        self.assertArrayAlmostEqual(
            compo.subcompos.geometric.bbox.bot_left,
            (3 + 10, 2 + 20)
            )

if __name__ == '__main__':
    unittest.main()




"""
transform.py: home to Transform class.

See the docstring of Transform for more information.
"""
from math import degrees

from copy import deepcopy

try:
    from typing import Self
except ImportError:
    # py3.10 and lower
    from typing_extensions import Self

import raimad as rai

class Transform:
    """Transformation: container for affine matrix."""

    _affine: 'rai.typing.Affine'

    def __init__(self) -> None:
        """Initialize a new Transform as an identity transform."""
        self.reset()

    def reset(self) -> None:
        """Reset this transform to be an identity transform."""
        self._affine = rai.affine.identity()

    def transform_xyarray(
            self,
            poly: 'rai.typing.Poly'
            ) -> 'rai.typing.Poly':  # TODO rename xyarray to polygon
        """
        Apply transformation to xyarray and return new transformed xyarray.

        Parameters
        ----------
        poly : rai.typing.Poly
            The Poly to transform

        Returns
        -------
        rai.typing.Poly
            The new, transformed, Poly
        """
        return rai.affine.transform_xyarray(self._affine, poly)

    def transform_point(
            self,
            point: 'rai.typing.PointLike'
            ) -> 'rai.typing.Point':
        """
        Apply this transform to a point, and return the transformed point.

        A Point (tuple of two floats) is always returned,
        even if a BoundPoint is passed in.

        Parameters
        ----------
        point : rai.typing.PointLike
            The point to transform.

        Returns
        -------
        rai.typing.Point
            The transformed point.
        """
        return rai.affine.transform_point(self._affine, point)

    def compose(self, transform: Self) -> Self:
        """
        Apply a different transform to this transform.

        Parameters
        ----------
        transform : Self
            The other transform to apply to this one

        Returns
        -------
        Self
            This transform is returned to allow method chaining.
        """
        if transform is not None:
            self._affine = rai.affine.matmul(transform._affine, self._affine)
        return self

    def get_translation(self) -> tuple[float, float]:
        """
        Return how much this transform translates the coordinate plane.

        Returns
        -------
        tuple[float, float]
            The first item is how much the plane is moved
            along the x axis.
            The second item is how much the plane is moved
            along the y axis.
        """
        return rai.affine.get_translation(self._affine)

    def get_rotation(self) -> float:
        """
        Return how much this transform rotates the coordinate plane.

        Returns
        -------
        float
            The angle by which this transform rotates the
            coordinate plane (radians, in the positive orientation)
        """
        return rai.affine.get_rotation(self._affine)

    def get_shear(self) -> float:
        """
        Return a how much this transform shears the coordinate plane.

        Returns
        -------
        float
            A number representing the shear.
        """
        return rai.affine.get_shear(self._affine)

    def get_scale(self) -> tuple[float, float]:
        """
        Return how much this transform scales the x and y axis.

        Returns
        -------
        tuple[float, float]
            The first float is the scaling factor along the x axis.
            The second float is the scaling factor along
            the y axis.
        """
        return rai.affine.get_scale(self._affine)

    def does_translate(self) -> bool:
        """
        Check whether this transform applies a translation.

        Returns
        -------
        bool
            True if the transform translates.
            False otherwise.
        """
        norm = rai.affine.norm(self.get_translation())
        return norm > rai.epsilon

    def does_rotate(self) -> bool:
        """
        Check whether this transform applies a rotation.

        Returns
        -------
        bool
            True if the transform rotates.
            False otherwise.
        """
        return abs(self.get_rotation()) > rai.epsilon

    def does_shear(self) -> bool:
        """
        Check whether this transform applies a shear.

        Returns
        -------
        bool
            True if the transform shears.
            False otherwise.
        """
        return abs(self.get_shear()) > rai.epsilon

    def does_scale(self) -> bool:
        """
        Check whether this transform applies a scale.

        Returns
        -------
        bool
            True if the transform scales.
            False otherwise.
        """
        scale_x, scale_y = self.get_scale()
        return abs(1 - scale_x) > rai.epsilon or abs(1 - scale_y) > rai.epsilon

    def __repr__(self) -> str:
        """
        Return string representation of transform.

        Returns
        -------
        str
            A representation of this transform.
        """
        does_translate = self.does_translate()
        does_rotate = self.does_rotate()
        does_shear = self.does_shear()
        does_scale = self.does_scale()

        move_x, move_y = self.get_scale()
        rotation = self.get_rotation()
        shear = self.get_shear()
        scale_x, scale_y = self.get_scale()

        if True not in {does_translate, does_rotate, does_shear, does_scale}:
            # Could also test if affine == identity
            return "<Identity Transform>"

        return ''.join((
            "<Transform ",
            f"Move ({move_x:+.2f}, {move_y:+.2f}) "
                if does_translate else '',
            f"Rotate {degrees(rotation):.2f}) "
                if does_rotate else '',
            f"Shear {shear:.2f} "
                if does_shear else '',
            f"Scale ({scale_x:.2f}, {scale_y:.2f})"
                if does_scale else '',
            ">",
            ))

    def copy(self) -> Self:
        """
        Duplicate transform.

        Returns
        -------
        Self
            The copied transform.
        """
        return deepcopy(self)

    # TODO typing.point
    # types defined in own files
    # then mergen in rai.typing
    # or just PointType, CompoClassType, etc
    def move(self, x: float, y: float) -> Self:
        """
        Move transform.

        Parameters
        ----------
        x : float
            Move this many units along x axis.
        y : float
            Move this many units along y axis.

        Returns
        -------
        Self
            This transform is returned to allow chaining methods.
        """
        self._affine = rai.affine.matmul(rai.affine.move(x, y), self._affine)
        return self

    def movex(self, x: float = 0) -> Self:
        """
        Move along x axis.

        Parameters
        ----------
        x : float
            Move this many units along x axis.

        Returns
        -------
        Self
            This transform is returned to allow chaining methods.
        """
        self._affine = rai.affine.matmul(rai.affine.move(x, 0), self._affine)
        return self

    def movey(self, y: float = 0) -> Self:
        """
        Move along y axis.

        Parameters
        ----------
        y : float
            Move this many units along y axis.

        Returns
        -------
        Self
            This transform is returned to allow chaining methods.
        """
        self._affine = rai.affine.matmul(rai.affine.move(0, y), self._affine)
        return self

    def scale(self, x: float, y: float | None = None) -> Self:
        """
        Scale a transform.

        Parameters
        ----------
        x : float
            Factor to scale by along the x axis
        y : float
            Factor to scale by along the y axis.
            If unspecified or None,
            use the x scale factor.

        Returns
        -------
        Self
            This transform is returned to allow chaining methods.
        """
        if y is None:
            y = x
        self._affine = rai.affine.matmul(rai.affine.scale(x, y), self._affine)
        return self

    def rotate(
            self,
            angle: float,
            x: float = 0,
            y: float = 0
            ) -> Self:
        """
        Rotate around a point.

        Parameters
        ----------
        angle : float
            Angle to rotate by, in radians.
        x : float
            Rotate around this point (x coordinate)
            default: origin
        y : float
            Rotate around this point (y coordinate)
            default: origin

        Returns
        -------
        Self
            This transform is returned to allow chaining methods.
        """
        self._affine = rai.affine.matmul(
            rai.affine.around(rai.affine.rotate(angle), x, y),
            self._affine
            )

        return self

    def hflip(self, x: float = 0) -> Self:
        """
        Flip (mirror) along horizontal axis.

        Parameters
        ----------
        x : float
            Flip around this horizontal line (y coordinate)

        Returns
        -------
        Self
            This transform is returned to allow chaining methods.
        """
        self._affine = rai.affine.matmul(
            rai.affine.around(rai.affine.scale(1, -1), 0, x),
            self._affine
            )
        return self

    def vflip(self, y: float = 0) -> Self:
        """
        Flip (mirror) along vertical axis.

        Parameters
        ----------
        y : float
            Flip around this vertical line (y coordinate)

        Returns
        -------
        Self
            This transform is returned to allow chaining methods.
        """
        self._affine = rai.affine.matmul(
            rai.affine.around(rai.affine.scale(-1, 1), y, 0),
            self._affine
            )
        return self

    def flip(self, x: float = 0, y: float = 0) -> Self:
        """
        Flip (mirror) along both horizontal and vertical axis.

        Parameters
        ----------
        x : float
            Flip around this point (x coordinate)
        y : float
            Flip around this point (y coordinate)

        Returns
        -------
        Self
            This transform is returned to allow chaining methods.
        """
        self._affine = rai.affine.matmul(
            rai.affine.around(rai.affine.scale(-1, -1), x, y),
            self._affine
            )
        return self




import unittest

import raimad as rai

from .utils import GeomsEqual, ArrayAlmostEqual

class TestPolys(GeomsEqual, ArrayAlmostEqual, unittest.TestCase):

    def test_ansec(self):
        same = [
            rai.AnSec(
                r1=10,
                r2=20,
                theta1=rai.eigthcircle,
                theta2=-rai.eigthcircle,
                ),
            rai.AnSec(
                r1=10,
                r2=20,
                theta1=rai.eigthcircle,
                theta2=-rai.eigthcircle,
                num_points=300,
                ),
            rai.AnSec.from_auto(
                r1=10,
                r2=20,
                theta1=rai.eigthcircle,
                theta2=-rai.eigthcircle,
                ),
            rai.AnSec.from_auto(
                r1=10,
                r2=20,
                theta1=-rai.eigthcircle,
                theta2=rai.eigthcircle,
                ),
            rai.AnSec.from_auto(
                r1=10,
                dr=10,
                theta1=-rai.eigthcircle,
                theta2=rai.eigthcircle,
                ),
            rai.AnSec.from_auto(
                r2=20,
                dr=10,
                theta1=-rai.eigthcircle,
                theta2=rai.eigthcircle,
                ),
            rai.AnSec.from_auto(
                r2=20,
                dr=10,
                theta1=-rai.eigthcircle,
                dtheta=rai.quartercircle,
                ),
            rai.AnSec.from_auto(
                r2=20,
                dr=10,
                theta2=rai.eigthcircle,
                dtheta=rai.quartercircle,
                ),
            rai.AnSec.from_auto(
                rmid=15,
                dr=10,
                theta2=-rai.eigthcircle,
                dtheta=-rai.quartercircle,
                ),
            rai.AnSec.from_auto(
                rmid=15,
                dr=10,
                thetamid=0,
                dtheta=rai.quartercircle,
                ),
            rai.AnSec.from_auto(
                rmid=15,
                dr=-10,
                thetamid=0,
                dtheta=-rai.quartercircle,
                ),
            ]

        for i, x in enumerate(same):
            open(f'/tmp/{i}.svg', 'w').write(rai.export_svg(x))

        i = 1
        for compo in same[1:]:
            print(i)
            i += 1
            self.assertArrayAlmostEqual(
                same[0].bbox.as_list(),
                compo.bbox.as_list(),
                epsilon=1
                )
            #self.assertGeomsEqual(
            #    same[0].geoms,
            #    compo.geoms
            #    )

        #geoms = [np.sort(ansec.geoms['root'][0], axis=0) for ansec in same]

        #for x in geoms:
        #    print(x)

        #self.assertTrue(np.allclose(geoms[0], geoms))

        # invalid
        with self.assertRaises(rai.err.AnSecError):
            # too many parameters for radius
            rai.AnSec.from_auto(
                r2=20,
                r1=10,
                dr=-10,
                theta1=-rai.eigthcircle,
                theta2=rai.eigthcircle,
                )

        with self.assertRaises(rai.err.AnSecError):
            # too many parameters for theta
            rai.AnSec.from_auto(
                r2=20,
                dr=10,
                theta2=rai.eigthcircle,
                theta1=0,
                dtheta=-rai.quartercircle,
                )

        with self.assertRaises(rai.err.AnSecError):
            # No radius specified
            rai.AnSec.from_auto(
                rmid=15,
                theta2=rai.eigthcircle,
                dtheta=rai.quartercircle,
                ),

        with self.assertRaises(rai.err.AnSecError):
            # No start angle specified
            rai.AnSec.from_auto(
                rmid=15,
                dr=5,
                dtheta=rai.semicircle,
                ),


if __name__ == '__main__':
    unittest.main()

